/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.sanskritlibrary.servlet;

import java.io.File;
import java.io.IOException;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletOutputStream;
import org.sanskritlibrary.Regex;
import java.io.FileInputStream;
import java.util.Hashtable;
import javax.servlet.http.Cookie;

/**
 *
 * @author rbunker
 */
public class FileFilter implements Filter
{

    public static Hashtable<String, String> contentTypes = new Hashtable<String, String>();
    // http://www.webmaster-toolkit.com/mime-types.shtml
    static
    {
        contentTypes.put("htm", "text/html; charset=UTF-8");
        contentTypes.put("html", "text/html; charset=UTF-8");
        contentTypes.put("pdf", "application/pdf");
        contentTypes.put("png", "image/png");
        contentTypes.put("jpg", "image/jpeg");
        contentTypes.put("jpeg", "image/jpeg");
        contentTypes.put("gif", "image/gif");
        contentTypes.put("js", "text/javascript; charset=UTF-8");
        contentTypes.put("css", "text/css; charset=UTF-8");
        contentTypes.put("xml", "text/xml; charset=UTF-8");
        contentTypes.put("rtf", "application/rtf");
        contentTypes.put("txt", "text/plain; charset=UTF-8");
    }
    private static final boolean debug = false;
    // The filter configuration object we are associated with.  If
    // this value is null, this filter instance is not currently
    // configured.
    private FilterConfig filterConfig = null;

    public FileFilter()
    {
    }

    public boolean loadFile(HttpServletResponse response, String filename) throws IOException
    {
        int extStart = filename.lastIndexOf('.');
        if (extStart != -1)
        {
            String ext = filename.substring(extStart + 1);
            String contentType = contentTypes.get(ext);
            response.setContentType(contentType);
        }
        File f = new File(filename);
        if (!f.exists()) {
            return false;
        } else {
            FileInputStream fis = new FileInputStream(filename);
            byte[] b = new byte[fis.available()];
            fis.read(b);
            fis.close();
            ServletOutputStream sos = response.getOutputStream();
            sos.write(b);
            sos.flush();
            sos.close();

            return true;
        }
    }

    public boolean doApacheProcessing(HttpServletRequest request, HttpServletResponse response,
            String filename, String ext) throws IOException
    {
        String apacheDirectory = (String) this.getFilterConfig().getServletContext().getAttribute("apacheDirectory");
        String subDirectory1 = (String) request.getSession().getAttribute("subDirectory1");
        String subDirectory2 = "";
        String fn = null;
        Regex re = new Regex(filename, "^/(.*/)([^/]+)$");
        if (re.find())
        {
            subDirectory2 = re.getGroup(1);
            if (subDirectory2.charAt(0) == '-')
            {
                subDirectory1 = subDirectory2.substring(1);
                subDirectory2 = "";
            }
            fn = re.getGroup(2);
        } else
        {
            subDirectory2 = "/";
            fn = filename.substring(1);
        }

        return loadFile(response, apacheDirectory + subDirectory1 + subDirectory2 + fn);
    }

    public void getTranscodedFile(HttpServletRequest request, HttpServletResponse response) throws IOException
    {
        String filename = null;
        Cookie[] cookies = request.getCookies();
        int len = cookies.length;
        for (int i = 0; i < len; i++)
        {
            Cookie cookie = cookies[i];
            if (cookie.getName().equals("transcodedFile"))
            {
                filename = cookie.getValue();
                break;
            }
        }

        if (filename != null)
        {
            String uploadDirectory = getFilterConfig().getServletContext().getInitParameter("uploadDirectory");
            loadFile(response, String.format("%s/%s", uploadDirectory, filename));
        }
    }

    public boolean doSpecialProcessing(HttpServletRequest request, HttpServletResponse response,
            String fn, String extension) throws IOException
    {
        String name = null;
        String ext = extension.toLowerCase();
        if (ext.equals("js")) {
            name = "jsDirectory";
        } else if (ext.equals("htm"))
        {
            name = "popupDirectory";
        } else if (ext.equals("css"))
        {
            name = "cssDirectory";
        } else if (ext.equals("jpeg") || extension.equals("jpg"))
        {
            name = "imageDirectory";
        } else if (ext.equals("gif"))
        {
            name = "imageDirectory";
        }

        if (name != null)
        {
            String dir = getFilterConfig().getServletContext().getInitParameter(name);
            String path = String.format("%s/%s.%s", dir, fn, extension);
            return loadFile(response, path);
        } else
        {
            return false;
        }
    }

    public void doFilter(ServletRequest request, ServletResponse response,
            FilterChain chain)
            throws IOException, ServletException
    {
        HttpServletRequest httpRequest = (HttpServletRequest) request;
        HttpServletResponse httpResponse = (HttpServletResponse) response;

        String url = httpRequest.getServletPath();
        if (url.startsWith("/download"))
        {
            getTranscodedFile(httpRequest, httpResponse);
            return;
        } else if (url.indexOf("preferences_htm") != -1)
        {
            doSpecialProcessing(httpRequest, httpResponse, "preferences", "htm");
            return;
        }
        boolean inApacheDirectory = ((HttpServletRequest) request).getSession().getAttribute("inApacheDirectory") != null;
        if ((inApacheDirectory || url.charAt(1) == '-') && url != null)
        {
            //System.out.format("In doFilter with %s\n", url);
            Regex re = new Regex(url, "\\.([^.]+)$");
            if (re.find())
            {
                String ext = re.getGroup(1).toLowerCase();
                if (ext.equals("pdf")
                        || ext.equals("htm")
                        || ext.equals("html")
                        || ext.equals("png")
                        || ext.equals("jpeg")
                        || ext.equals("jpg")
                        || ext.equals("gif"))
                {
                    if (!doApacheProcessing(httpRequest, httpResponse, url, ext))
                        chain.doFilter(request, response);
                } else
                    chain.doFilter(request, response);
            } else // There is no extension, assume that it is referencing a Servlet and that we are back in Tomcat land
            {
                HttpServletRequest rq;
                ((HttpServletRequest) request).getSession().removeAttribute("inApacheDirectory");
                ((HttpServletRequest) request).getSession().setAttribute("subDirectory1", "/");
                chain.doFilter(request, response);
            }
        } else
            chain.doFilter(request, response);
        }

    /**
     * Return the filter configuration object for this filter.
     */
    public FilterConfig getFilterConfig()
    {
        return (this.filterConfig);
    }

    /**
     * Set the filter configuration object for this filter.
     *
     * @param filterConfig The filter configuration object
     */
    public void setFilterConfig(FilterConfig filterConfig)
    {
        this.filterConfig = filterConfig;
    }

    /**
     * Destroy method for this filter
     */
    public void destroy()
    {
    }

    /**
     * Init method for this filter
     */
    public void init(FilterConfig filterConfig)
    {
        this.filterConfig = filterConfig;
    }
}
