//http://www.sqlite.org/download.html
// help@brown.edu or Elli Mylonas <Elli_Mylonas@Brown.EDU>
package org.sanskritlibrary.servlet;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.Cookie;
import javax.servlet.ServletInputStream;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import org.sanskritlibrary.Cacher;

import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.TextFormatter;
import org.sanskritlibrary.TextStructure;

public class SanskritLibraryServlet extends HttpServlet implements ResourceInputStream {

    private static int debugCount = 0;
    private String lastError = null;        // TODO all threads will compete for this. Use exception instead

    protected Hashtable<String, String> createArgHash(HttpServletRequest request, String... args) {
        Hashtable<String, String> h = new Hashtable<String, String>();
        for (String arg : args) {
            h.put(arg, request.getParameter(arg));
        }
        return h;
    }

    protected void writeResult(HttpServletResponse response, String result, String... contentType) throws IOException {
        response.setCharacterEncoding("UTF-8");
        String contentTypeString = null;
        if (contentType.length == 0) {
            contentTypeString = "text/text";
        } else {
            contentTypeString = contentType[0];
        }

        response.setContentType(contentTypeString);
        OutputStream os = response.getOutputStream();
        OutputStreamWriter osw = new OutputStreamWriter(os, "UTF-8");
        PrintWriter pw = new PrintWriter(osw);

        pw.print(result);
        pw.flush();
        pw.close();
    }

    protected String getCookieValue(HttpServletRequest request, String name) {
        String value = null;
        if (request.getCookies() != null) {
            for (Cookie c : request.getCookies()) {
                if (c.getName().equals(name)) {
                    value = c.getValue();
                    break;
                }
            }
        }
        return value;
    }

    protected String getCookieValue(HttpServletRequest request, String name, String dflt) {
        String result = getCookieValue(request, name);
        if (result == null) {
            result = dflt;
        }

        return result;
    }

    protected String getParameterValue(HttpServletRequest request, String name, String dflt) {
        String result = request.getParameter(name);
        if (result == null) {
            result = dflt;
        }

        return result;
    }

    public TransformMap getTransformMap(String map) {
        if (map == null || map.length() == 0) {
            return null;
        }

        Hashtable<String, FSM[]> fsmh =
                (Hashtable<String, FSM[]>) this.getServletContext().getAttribute("fsmh");
        if (fsmh == null) {
            fsmh = new Hashtable<String, FSM[]>();
            this.getServletContext().setAttribute("fsmh", fsmh);
        }
        try {
            TransformMap tm = new TransformMap(map, this, fsmh);
            return tm;
        } catch (Exception ex) {
            lastError = ex.getMessage();
            return null;
        }
    }

    public String transformStringRegex(String text, String map, String regex) {
        return transformStringRegex(null, text, map, regex);
    }

    public String transformStringRegex(Cacher cacher, String text, String map, String regex) {
        if (map == null || map.length() == 0) {
            return text;
        }
        Hashtable<String, FSM[]> fsmh =
                (Hashtable<String, FSM[]>) this.getServletContext().getAttribute("fsmh");
        if (fsmh == null) {
            fsmh = new Hashtable<String, FSM[]>();
            this.getServletContext().setAttribute("fsmh", fsmh);
        }

        try {
            debugCount++;
            TransformMap tm = new TransformMap(map, this, fsmh);
            if (regex == null) {
                return WebServices.transformString(cacher, text, tm);
            } else {
                return WebServices.transformString(cacher, text, tm, regex);
            }
        } catch (Exception ex) {
            lastError = ex.getMessage();
            return null;
        }
    }

    public String transformString(Cacher cacher, String text, String map) {
        return transformStringRegex(cacher, text, map, null);
    }

    public String transformString(String text, String map) {
        return transformString(null, text, map);
    }

    public String transformUnicodeString(String result, String map, int low, int high) {
        return transformUnicodeString(null, result, map, low, high);
    }

    public String transformUnicodeString(Cacher cacher, String result, String map, int low, int high) {
        // collect devanagari characters and convert them to target encoding
        StringBuilder sb = new StringBuilder();
        StringBuilder deva = new StringBuilder();
        for (int i = 0; i < result.length(); i++) {
            char ch = result.charAt(i);
            if ((int) ch >= low && (int) ch <= high) {
                deva.append(ch);
            } else {
                if (deva.length() > 0) {
                    sb.append(transformString(cacher, deva.toString(), map));
                    deva.setLength(0);
                }
                sb.append(ch); // append non-devanagari character
            }
        }

        return sb.toString();
    }

    public InputStream[] getInputStreams(String sourceEncoding, String targetEncoding) {
        String datadir = getServletContext().getInitParameter("dataDirectory");
        String sep = "";
        if (!datadir.endsWith("/")) {
            sep = "/";
        }
        //String transcoderDirectory = getServletContext().getInitParameter("transcoderDirectory");
        String transcoderDirectory = String.format("%s%stranscoders", datadir, sep);
        String[] fileNames = FSM.createFileNames(transcoderDirectory, sourceEncoding, targetEncoding);
        InputStream[] iss = new InputStream[fileNames.length];
        int i = 0;

        for (String name : fileNames) {
            try {
                iss[i] = new FileInputStream(name);
            } catch (Exception ex) {
                iss[i] = null;
                System.out.format("Error in getInputStreams: %s\n", ex.getMessage());
            }
            i++;
        }


        return iss;

    }

    // This is the old way of doing it when the resources were in the war file
    public InputStream[] getInputStreamsResources(String sourceEncoding, String targetEncoding) {
        String[] fileNames = FSM.createFileNames("/WEB-INF/classes/transcoder/", sourceEncoding, targetEncoding);
        InputStream[] iss = new InputStream[fileNames.length];
        int i = 0;

        for (String name : fileNames) {
            iss[i++] = getServletContext().getResourceAsStream(name);
        }

        return iss;

    }

    // returns a hash table with all HTTP headers except Cookie
    public Hashtable<String, String> createHashtable(HttpServletRequest request) {
        Enumeration en = request.getHeaderNames();
        Hashtable<String, String> h = new Hashtable<String, String>();
        while (en.hasMoreElements()) {
            String name = (String) en.nextElement();
            if (!name.toLowerCase().equals("cookie")) {
                h.put(name, request.getHeader(name));
            }
        }

        return h;
    }

    public String mapPageDisplayEncoding(HttpServletRequest request) {
        String encoding = this.getCookieValue(request, "serverOptions", "SLP1");
        String filter = "";

        if (encoding.equals("deva")) {
            filter = "SktDevaUnicode";
        } else if (encoding.equals("roman")) {
            filter = "SktRomanUnicode";
        }

        return filter;
    }

    public String getViewingType(HttpServletRequest request) // what user sees when he types, hence the format of what is sent
    {
        String inputType = getCookieValue(request, "inputType", "phonetic");
        if (inputType.equals("phonetic")) {
            // we know that user was typing some phonetic encoding such as SLP1 but
            // we need to know how the typing was actually displayed.
            inputType = getCookieValue(request, "viewAs", "slp1");
            if (inputType.equals("phonetic")) // but if phonetic encoding was displayed, then get phoneticInput
            {
                inputType = getCookieValue(request, "phoneticInput", "slp1");
            }
        }
        return inputType;
    }

    public String getOutputType(HttpServletRequest request) // the format of what is sent if a link is clicked, i.e., the user did not type anything, the format is whatever the server sent.
    {
        String serverOptions = getCookieValue(request, "serverOptions", "deva");
        if (serverOptions.equals("phonetic")) {
            serverOptions = getCookieValue(request, "phonetic", "slp1");
        }

        return serverOptions;
    }

    public String convertToSLP1(Cacher cacher, String sourceType, String s) {
        if (!sourceType.equals("slp1")) {
            return transformString(cacher, s, String.format("*:%s->slp1", sourceType));
        } else {
            return s;
        }
    }

    // Assume response type is ..., Creates a hash table of parameters including
    // uploaded files.
    public Hashtable<String, String> getParameters(HttpServletRequest request)
            throws IOException {
        return getParameters(request, null, false);
    }

    public Hashtable<String, String> getParameters(HttpServletRequest request, 
            Hashtable<String, Boolean> binary, boolean returnBytes)
            throws IOException {
        Hashtable<String, String> params = new Hashtable<String, String>();
        String type = request.getContentType();
        if (type.contains("application/x-www-form-urlencoded")) {
            Enumeration en = request.getParameterNames();
            while (en.hasMoreElements()) {
                String name = (String) en.nextElement();
                params.put(name, request.getParameter(name));
            }
            return params;
        } else if (!type.contains("multipart/form-data")) {
            lastError = "Only multipart/form-data is supported";
            return null;
        } else {
            int len = request.getContentLength();
            ServletInputStream sis = request.getInputStream();
            byte[] bfr = new byte[len];
            String delim = null;
            int delimlen = 0;
            int currline = 0;
            boolean readcontent = false;
            String name = null;
            String filename = null;
            int i = 0;
            int j = 0;
            while (i < bfr.length) {
                byte b = (byte) sis.read();
                i++;
                if (b == 0x0d) {
                    b = (byte) sis.read();
                    i++;
                    if (b != 0x0a) {
                        lastError = "Corrupted request. 0x0d not followed by 0x0a";
                        return null;
                    }
                    String line = new String(bfr, 0, j, "UTF-8");
                    j = 0;
                    currline = 0;

                    if (delim == null) // first line is the line delimeter to use
                    {
                        delim = line;
                        delimlen = line.length();
                    } else {
                        if (line.startsWith("Content-Disposition:")) {
                            Regex re = new Regex(line.substring(20), "([^;]+)");
                            boolean found = false;
                            while (re.find()) {
                                found = true;
                                String s = re.getGroup(1).trim();
                                if (s.startsWith("name=")) {
                                    name = s.substring(5);
                                    char ch = name.charAt(0);
                                    if (ch == '"' || ch == '\'') {
                                        name = name.substring(1, name.length() - 1);
                                    }
                                } else if (s.startsWith("filename=")) {
                                    filename = s.substring(9);
                                    char ch = filename.charAt(0);
                                    if (ch == '"' || ch == '\'') {
                                        filename = filename.substring(1, filename.length() - 1);
                                    }
                                }
                            }
                            if (!found) {
                                lastError = "Error parsing Content-Disposition line";
                                return null;
                            }
                        } else if (line.startsWith("Content-Type:")) {
                            /*
                            if (!line.contains("text/plain"))
                            {
                            lastError = String.format("Content-Type is %s, however, the uploaded file must be of type text/plain", line);
                            //return null;
                            }
                             */
                        } else if (line.length() == 0) {
                            readcontent = true;			// this will detect delimiter
                        }
                    }

                    if (readcontent) // read either value of parameter or file contents
                    {
                        readcontent = false;
                        while (i < bfr.length) {
                            b = (byte) sis.read();
                            i++;
                            if (b == 0x0d) {
                                bfr[j++] = b;
                                b = (byte) sis.read();
                                i++;
                                if (b == 0x0a) // possibly have reached the delimiter line
                                {
                                    bfr[j++] = b;
                                    String sd = new String(bfr, currline, j - currline, "UTF-8");
                                    if (sd.length() >= delimlen
                                            && sd.substring(0, delimlen).equals(delim)) {
                                        if (filename != null && filename.length() > 0) {
                                            params.put(name + ".filename", filename);
                                        }

                                        if (returnBytes) {
                                            byte[] bb = new byte[currline - 2];
                                            for (int k = 0; k < bb.length; k++) {
                                                bb[k] = bfr[k];
                                            }
                                            request.setAttribute(filename, bb);
                                        } else {
                                            if (binary != null && binary.get(name) != null) {
                                                byte[] bb = new byte[currline - 2];
                                                for (int k = 0; k < bb.length; k++) {
                                                    bb[k] = bfr[k];
                                                }
                                                request.setAttribute(name, bb);
                                            } else {
                                                String value = new String(bfr, 0, currline, "UTF-8");

                                                value = value.substring(0, value.length() - 2);	// remove last \r\n

                                                params.put(name, value);
                                            }
                                        }

                                        name = null;
                                        filename = null;
                                        j = 0;
                                        break;
                                    }
                                    currline = j;		// start of next line
                                } else {
                                    bfr[j++] = b;
                                }
                            } else {
                                bfr[j++] = b;
                            }
                        }
                    }
                } else {
                    bfr[j++] = b;
                }
            }		// end of for loop
            return params;
        }
    }

    protected void saveViewState(HttpServletRequest request, Hashtable<String, String> h, String... params) {
        int debug = 0;
        StringBuilder sb = new StringBuilder();
        for (String param : params) {
            String value = null;
            String id = "vs_" + param;
            if (h != null) {
                value = h.get(id);
            } else {
                value = request.getParameter(param);
            }
            if (value == null) {
                value = "";
            }
            sb.append(
                    String.format("<input id='%s' name='%s' type='hidden' value='%s'/>\n",
                    id, id, TextFormatter.charToEntity(value)));
        }
        request.setAttribute("viewstate", sb.toString());
    }

    protected TextStructure getTextStructure(HttpServletRequest request, String tsfn) throws Exception {
        TextStructure ts = null;
        String currentTS = (String) request.getSession().getAttribute("tsfn");
        if (currentTS != null && currentTS.equals(tsfn)) {                  // is this TextStructure in memory?
            ts = (TextStructure) request.getSession().getAttribute("ts");   // use it
        } else {    // otherwise load it from disk
            request.getSession().setAttribute("tsfn", tsfn);
            FileInputStream fis = new FileInputStream(tsfn);
            ObjectInputStream ois = new ObjectInputStream(fis);
            ts = (TextStructure) ois.readObject();
            ois.close();
            request.getSession().setAttribute("ts", ts);
        }

        return ts;
    }

    protected void forward(String name, HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        RequestDispatcher rd = request.getRequestDispatcher(name);
        rd.forward(request, response);
    }

    protected void forward(String name, HttpServletRequest request, HttpServletResponse response,
            Hashtable<String, String> h, String... params)
            throws ServletException, IOException {
        saveViewState(request, h, params);
        request.setAttribute("helpfile", "helpNotAvailable.html");
        RequestDispatcher rd = request.getRequestDispatcher(name);
        rd.forward(request, response);
    }

    protected void createAttributes(HttpServletRequest request, String[] params) {
        for (String param : params) {
            request.setAttribute(param, request.getParameter(param));
        }
    }

    protected String getError() {
        return lastError;
    }
}
